#include "IR.hpp"
#include "AST.hpp"
#include "GRCPrinter.hpp"
#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <stdio.h>
#include <set>

using namespace AST;
using namespace std;
using namespace GRCDot;

class CUTVAR{
public:
  set<GRCNode *> forks_attach;
  set<GRCNode *> parents_cut;
  SignalSymbol *variable;
  int value;

  CUTVAR(GRCNode *fn, GRCNode *pn, SignalSymbol *variable, int value): variable(variable), value(value)
  {
    forks_attach.insert(fn);
    parents_cut.insert(pn);
  }
};

class PDG2CCFG{
  GRCNode *top;
  SymbolTable *sigs;
  CFGmap &refmap; 
  GRCNode *toppdg;
  EnterGRC *entergrc;

  int newsig_counter; //to save the new signals number
  char *newsig_prefix; //unique prefix of new signals

  set<GRCNode *> visited;
  vector<GRCNode *> topo;

  //for ordering
  map<GRCNode *, set<GRCNode *> > dpsnodes_set;
  map<GRCNode *, vector<GRCNode *> > dpsnodes_vec;
  map<GRCNode *, vector<int> > dpsnodes_brnumb;

  //for cuting
  map<GRCNode *, GRCNode *> active; //currently active brunch of forks

  map<GRCNode *, CUTVAR *> att_point; //attaching point

  map<GRCNode *, vector<GRCNode *> > conn_succ; //saved connection - successors
  map<GRCNode *, vector<GRCNode *> > conn_pred; //saved connection - predecessors
  map<GRCNode *, vector<vector<GRCNode *>::iterator> > null_ch; //for nodes having null children
  map<GRCNode *, GRCNode *> replaced; //new emitters -> replaced nodes

  int debug;

public:
  PDG2CCFG(GRCNode *top, SymbolTable *sigs, CFGmap &refmap): top(top), sigs(sigs), refmap(refmap)
  {
    debug = 0;

    entergrc = dynamic_cast<EnterGRC*>(top); 
    assert(entergrc);
    assert(entergrc->successors.size() == 1);
    toppdg = dynamic_cast<GRCNode*>(entergrc ->successors[0]); 
    assert (toppdg);

    newsig_counter = 0;
    newsig_prefix = "_DPSCUT_VAR";
  }
  
  void excute()
  { 
    //later consider also ordering switches' children   
    orderingDFS(toppdg);

    visited.clear();
    topoDFS(toppdg);
    //printTopo();

    cutTopo();

    removeConn();
    recoverConn();
  }

  void removeConn()
  {
    set<GRCNode *>::iterator i;
    vector<GRCNode *>::iterator j;

    for (i = visited.begin(); i != visited.end(); i++){
      assert(*i);

      for (j = (*i)->successors.begin(); j != (*i)->successors.end(); j++)
	if (!(*j))
	  null_ch[*i].push_back(j);

      (*i)->successors.clear();
      (*i)->predecessors.clear();
    }
  }

  void recoverConn()
  {
    set<GRCNode *>::iterator i;
    vector<GRCNode *>::iterator j;
    
    if (debug) cerr<<"recovering connection.........\n";
    for (i = visited.begin(); i != visited.end(); i++){      
      assert(*i);
      if (debug) cerr<<"N"<<refmap[*i]<<":\n";
      if (conn_succ.count(*i)>0){
	if (debug) cerr<<"succ: ";
	for (j = conn_succ[*i].begin(); j != conn_succ[*i].end(); j++){
	  if(debug) cerr<<refmap[*j]<<" ";
	  (*i)->successors.push_back(*j);
	}

	if (null_ch.find(*i) != null_ch.end()){
	  vector<vector<GRCNode *>::iterator>::iterator k;
	  for (k = null_ch[*i].begin(); k != null_ch[*i].end(); k++){
	    (*i)->successors.insert(*k, NULL);
	  }
	}

	if (debug) cerr<<"\n";
      }
      assert(conn_pred.count(*i)>0);
      if(debug) cerr<<"pre: ";
      for (j = conn_pred[*i].begin(); j != conn_pred[*i].end(); j++){
	assert(*j);
	if(debug) cerr<<refmap[*j]<<" ";
	(*i)->predecessors.push_back(*j);
      }
      if(debug) cerr<<"\n";
    }
  }

  void cutTopo()
  {
    vector<GRCNode *>::iterator i;

    set<GRCNode *> fps; //most recent fork parents set
    set<GRCNode *>::iterator vi;

    for (i = topo.end()-1; i >= topo.begin(); i--){

      if(debug)
	cerr<<"topo visiting N"<<refmap[*i]<<"\n";
      
      //find the closest-fork-parents set for *i
      fps = mostRFPV(*i);

      if (fps.size() > 0){
	set<GRCNode *> cut;
	set<GRCNode *>::iterator cut_i;

	for (vi = fps.begin(); vi != fps.end(); vi++){

	  if(att_point.find(*i) != att_point.end()){
	    if (att_point[*i]->parents_cut.find(*vi) != att_point[*i]->parents_cut.end())
	      continue; //it was cut
	  }

	  if(debug) cerr<<"for most recent fp "<<refmap[*vi]<<"\n";
	  	  
	  GRCNode *fork_par = *vi;
	  GRCNode *by = brunchBelong(*i, fork_par); // due to pdg's defination, by has only 1 value
	  assert(by); //assert such a by exist

	  if(debug) cerr<<refmap[*i]<<" belongs to brunch "<<refmap[by]<<"\n";

	  if (active.find(fork_par) == active.end())
	    active[fork_par] = by;
	  
	  if (active[fork_par] != by){ 
	    //cerr<<"switch from "<<refmap[fork_par]<<"'s br "<<refmap[active[fork_par]]
	    //<<" to "<<refmap[by]<<"\n";
	    cutDown(fork_par, active[fork_par], cut); //check & cut down when find unfinished brunches
	    active[fork_par] = by;
	  }
	}

	for (cut_i = cut.begin(); cut_i != cut.end(); cut_i++){
	  checkParentCut(*cut_i, cut);
	}
      }
      addBack(*i);  
      if(debug) 
	cerr<<"finished topo visiting N"<<refmap[*i]<<"\n";    
    }
  }

  //for each n, it can only be added back once
  void addBack(GRCNode *n)
  {
    vector<GRCNode *>::iterator i;
    set<GRCNode *>::iterator j;

    if(debug)
      cerr<<"TO add back "<<refmap[n]<<"\n";

    if (att_point.find(n) != att_point.end()){
      CUTVAR *cv = att_point[n];      
      GRCNode *tester = newTester(cv->variable);
      visited.insert(tester);//visited-set should include all nodes in PDG
      if(debug) cerr<<"try to newLink tester\n";
      newLink(tester,NULL);
      newLink(tester,n);

      //all parents cut should be removed - not attach back
      for (j = cv->forks_attach.begin(); j != cv->forks_attach.end(); j++){
	if(debug) cerr<<"try to newLink tester2\n";
	newLink(*j, tester);
      }
      //others, attach them back
      for(i = n->predecessors.begin(); i != n->predecessors.end(); i++){
	if(cv->parents_cut.find(*i) == cv->parents_cut.end()){
	  if(debug) cerr<<"1try to link "<<refmap[*i]<<" - "<< refmap[n]<<"\n";
	  Link(*i,n);
	}
      }
    }
    else{
      for(i = n->predecessors.begin(); i != n->predecessors.end(); i++){
	if(debug) cerr<<"2try to link "<<refmap[*i]<<" - "<< refmap[n]<<"\n";
	Link(*i,n);
      }
    }
    if (n->successors.size()>0)
      conn_succ[n]; //access it to make it appear in the map but do nothing with it
  }

  //for nodes ch those were not in par->successors in pdg
  void newLink(GRCNode *par, GRCNode *ch)
  {
    assert(par);

    if(ch)
      conn_pred[ch].push_back(par);

    conn_succ[par].push_back(ch);

  }

  //for nodes which were in par->successors in pdg
  void Link(GRCNode *par, GRCNode *ch)
  {
    Link(par, ch, ch);
  }

  void Link(GRCNode *par, GRCNode *ch, GRCNode *newch)
  {
    assert(par);

    /*if (att_point.find(ch) != att_point.end()){
      if (att_point[ch]->parents_cut.count(par) > 0){
	cerr<<"not linked\n";
	return; //cut, dont add back
      }
      }*/


    if (conn_succ[par].size() == 0){
      conn_succ[par].push_back(newch);
      conn_pred[newch].push_back(par);
      return;
    }

    vector<GRCNode *>::iterator conn_p;
    int succ_p, succ_p_pre, pos_p;
    bool find;
    
    succ_p = 0;
    succ_p_pre = 0;
    pos_p = 0;
    conn_p = conn_succ[par].begin();

    if(debug) {
      cerr<<refmap[par]<<": before link, the conn has: ";
      printVec(conn_succ[par]);
    }
    pos_p = moveP(pos_p, par->successors, ch);    
    succ_p = moveP(succ_p, par->successors, *conn_p);    
    if(debug) cerr<<"looking for ch = "<<refmap[ch]<<", conn = "<<refmap[*conn_p]<<"\n";
    if(debug) cerr<<"first, pos_p = "<<pos_p<<", succ_p = "<<succ_p<<"\n";

    find = false;
    while ((!find) && (conn_p != conn_succ[par].end())){
      if (succ_p < 0){
	//if it's a new tester add to the successors
	succ_p = succ_p_pre;  //restore succ_p
	conn_p++;
	succ_p = moveP(succ_p, par->successors, *conn_p);
      }
      else{
	find = true;
	succ_p_pre = succ_p;
      }
    }
    
    while ((pos_p > -1) && (conn_p != conn_succ[par].end())){
      if (pos_p < succ_p){
	if(debug) cerr<<"insert "<<refmap[newch]<<" ("<<refmap[ch]<<") before "<<refmap[*conn_p]<<"\n";
	vector<GRCNode *>::iterator i_inserted = conn_succ[par].insert(conn_p, newch);
	conn_pred[newch].push_back(par);
	conn_p = (i_inserted)+1;
	pos_p++;
	pos_p = moveP(pos_p, par->successors, ch);

      }
      else{
	conn_p++;
	succ_p++;
	succ_p = moveP(succ_p, par->successors, *conn_p);
	find = false;
	while ((!find) && (conn_p != conn_succ[par].end())){
	  if (succ_p < 0){
	    //if it's a new tester add to the successors
	    succ_p = succ_p_pre;  //restore succ_p
	    conn_p++;
	    succ_p = moveP(succ_p, par->successors, *conn_p);
	  }
	  else{
	    find = true;
	    succ_p_pre = succ_p;
	  }
	}
      }
    }

    if (conn_p >= conn_succ[par].end()){
      conn_succ[par].push_back(newch);
      conn_pred[newch].push_back(par);
    }
    if(debug){
      cerr<<"after link, the conn has: ";
      printVec(conn_succ[par]);
    }
  }

  void printVec(vector<GRCNode *> invec)//for debug only
  {
    vector<GRCNode *>::iterator i;
    for (i = invec.begin(); i != invec.end(); i++){
      cerr<<refmap[*i]<<" ";
    }
    cerr<<"\n";
  }

  int moveP(int starti, vector<GRCNode *> inset, GRCNode *n)
  {
    GRCNode *lookingfor;

    if (replaced.find(n) != replaced.end()) //if n is an emitter which replaced another node
      lookingfor = replaced[n]; //instead of looking for n, looking for the replaced node
    else
      lookingfor = n;

    int x = 0;
    int RET = starti;
    while ((inset[RET] != lookingfor) && (RET < (int)inset.size())){
      x++;
      RET++;
    }

    if (RET >= (int)inset.size())
      return -1;// the end of vector
    return RET; // find n at RET position
  }

  //not only cut myself, but also cut the nodes under & have been cut from above(reversed)
  void cutDown(GRCNode *fork, GRCNode *brunch, set<GRCNode *> &cut)
  {
    set<GRCNode *> black;

    if(debug) cerr<<"begin CUTDFS under "<<refmap[brunch]<<"\n";
    cutDFS(brunch, fork, cut, black);
  }

  void cutDFS(GRCNode *n, GRCNode *attn, set<GRCNode *> &cut, set<GRCNode *> &black)
  {
    if (!n)
      return;        

    //cerr<<"visiting "<<refmap[n]<<"\n";
    if (black.find(n) != black.end()){
      //cerr<<"visited before\n";
      return;
    }
    black.insert(n);
    
    if (conn_succ.find(n) != conn_succ.end()){
      //cerr<<refmap[n]<<" has successors\n";
      if (!findAllNNullSucc(n)){
	//it means, n has successors, but some are not connected -CUT
	vector<GRCNode *>::iterator i;

	//cerr<<"but not connected--CUT!!!!!!!!!! under N"<<refmap[n]<<"\n";

	for (i = n->successors.begin(); i != n->successors.end(); i++){
	  if (!findInVec(*i, conn_succ[n])){
	    cut.insert(*i);
	    //cerr<<"CUT!! before "<<refmap[*i]<<" from path "<<refmap[n]<<"\n";
	    if(att_point.find(*i) == att_point.end()){
	      att_point[*i] = new CUTVAR(attn, n, newVar(),1);
	    }
	    else{
	      att_point[*i]->forks_attach.insert(attn);
	      att_point[*i]->parents_cut.insert(n);
	    }

	    GRCNode *ne = newEmitter(att_point[*i]->variable);
	    replaced[ne] = *i;
	    visited.insert(ne); //visited-set should include all nodes in PDG
	    Link(n,*i,ne);

	    //checkParentCut(*i, cut);
	  }
	  else
	    cutDFS(*i, attn, cut,  black);
	}
	return;
      }
      else {
	//cerr<<"visiting "<<refmap[n]<<"'s children\n";
	for (vector<GRCNode *>::iterator i = conn_succ[n].begin();
	     i != conn_succ[n].end(); i++){
	  cutDFS(*i, attn, cut,  black);
	}
      }
    }
    //else: n has no null children

    //cerr<<"finished visiting "<<refmap[n]<<"\n";
  }

  bool findAllNNullSucc(GRCNode *n)
  {
    vector<GRCNode *>::iterator i,j;
    vector<GRCNode *> conn = conn_succ[n];
    vector<GRCNode *> succ = n->successors;

    for (i = succ.begin(); i != succ.end(); i++){
      if(!*i)
	continue;
      if(att_point.find(*i) != att_point.end()){
	if (att_point[*i]->parents_cut.find(n) != att_point[*i]->parents_cut.end())
	  continue;
      }
      j = conn.begin(); 

      while ((j != conn.end()) && (*j != *i))
	j++;
      if (j == conn.end())
	return false;
    }
    return true;

  }

  bool findInVec(GRCNode *n, vector<GRCNode *> invec)
  {
    vector<GRCNode *>::iterator i;
    for (i = invec.begin(); i != invec.end(); i++)
      if (*i == n)
	return true;
    return false;
  }

  void checkParentCut(GRCNode *n, set<GRCNode *> cut)
  {
    set<GRCNode *> black;
    vector<GRCNode *>::iterator i;
    map<GRCNode *, bool> cutabove;
    
    for (i = n->predecessors.begin(); i != n->predecessors.end(); i++){
      if (checkPCutDFS(*i, cut, black, cutabove)){
	assert(att_point.find(n) != att_point.end());
	assert(cut.find(n) != cut.end());
	att_point[n]->parents_cut.insert(*i);
	//cerr<<"also CUT!! "<<refmap[n]<<" from path "<<refmap[*i]<<"\n";

	GRCNode *ne = newEmitter(att_point[n]->variable);
	replaced[ne] = n;
	visited.insert(ne); //visited-set should include all nodes in PDG
	Link(*i,n,ne);
      }
    }
  }
  
  bool checkPCutDFS(GRCNode *n, set<GRCNode *> cut, set<GRCNode *> &black, 
		    map<GRCNode *, bool> &cutabove)
  {
    vector<GRCNode *>::iterator i;

    if (!n)
      return false;

    if(black.find(n) != black.end()){
      assert(cutabove.find(n) != cutabove.end());
      return cutabove[n];
    }
    black.insert(n);

    if(cut.find(n) != cut.end()){
      cutabove[n] = true;
      return true;
    }
    
    for (i = n->predecessors.begin(); i != n->predecessors.end(); i++){
      if (checkPCutDFS(*i, cut, black, cutabove))
	return true;
    }

    cutabove[n] = false;
    return false;
  }

  //find which brunch ch belongs to
  GRCNode *brunchBelong(GRCNode *ch, GRCNode *par)
  {
    set<GRCNode *> black;

    return (brSearchDFS(ch, par, black));
  }

  GRCNode *brSearchDFS(GRCNode *n, GRCNode *par, set<GRCNode *> &black)
  {
    vector<GRCNode *>::iterator i;

    if (black.count(n)){
      return NULL;
    }

    black.insert(n);

    for (i = n->predecessors.begin(); i != n->predecessors.end(); i++){
      if (*i == par)
	return n;
      else {
	GRCNode *tmp = brSearchDFS(*i, par, black);
	if (tmp)
	  return tmp;
      }
    }

    return NULL;
  }

  //find the most recent fork parents those are visiting
  set<GRCNode *> mostRFPV(GRCNode *n) 
  {
    set<GRCNode *> black;

    return getForkParents(n, black);

  }

  set<GRCNode *> getForkParents(GRCNode* n, set<GRCNode *> &black)
  {
    vector<GRCNode *>::iterator i;
    set<GRCNode *> RET;
    
    if (black.count(n) > 0)
      return RET;
    black.insert(n);

    for (i = n->predecessors.begin(); i != n->predecessors.end(); i++){
      if (dynamic_cast<Fork *>(*i)){
	RET.insert(*i); //once meet a fork parent, stop reverse-dfs
      }
      else{
	set<GRCNode *> furtherps = getForkParents(*i, black);
	RET.insert(furtherps.begin(), furtherps.end());
      }
    }
    
    return RET;
  }

  void printTopo()
  {
    vector<GRCNode *>::iterator i;

    cerr<<"TOPO: ";
    for (i = topo.end()-1; i >= topo.begin(); i--)
      cerr<<refmap[*i]<<" ";
    cerr<<"\n";
  }

  void topoDFS(GRCNode *n)
  {
    vector<GRCNode *>::iterator i;
  
    if(!n) return;
    
    if(visited.count(n)>0) 
      return;

    visited.insert(n);

    for (i = n->successors.begin(); i != n->successors.end(); i++){
      topoDFS(*i);
    }

    if (n->dataSuccessors.size()>0){
      for (i = n->dataSuccessors.begin();
	   i != n->dataSuccessors.end(); i++)
	topoDFS(*i);
    }
    
    topo.push_back(n);

    return;
  }

  SignalSymbol *newVar()
  {
    char new_sig_name[32];

    sprintf(new_sig_name, "%s%d", newsig_prefix, newsig_counter);
    newsig_counter++;
    SignalSymbol *newsym = new SignalSymbol(new_sig_name,NULL,
					    SignalSymbol::Input,NULL,NULL);
    sigs->enter(newsym);
    return newsym;
  }

  GRCNode *newEmitter(SignalSymbol *sig)
  {
    Emit *em = new Emit(sig, NULL);
    return (new Action(em));
  }

  GRCNode *newTester(SignalSymbol *sig)
  {
    LoadSignalExpression *lde = new LoadSignalExpression(NULL, sig);
    return (new Test(NULL, lde));
  }

  void orderingDFS(GRCNode *n)
  {
    bool isfork = false;
    int succnumb;

    if (!n) 
      return;
    if (visited.count(n) >0)
      return;

    visited.insert(n);    
    
    if ((dynamic_cast<Fork *>(n)))
      isfork = true;
    
    succnumb = 0;
    for (vector<GRCNode *>::iterator i = n->successors.begin();
	 i != n->successors.end(); i++){
      orderingDFS(*i);
      if (dpsnodes_set.count(*i)>0){
	dpsnodes_set[n].insert(dpsnodes_set[*i].begin(), dpsnodes_set[*i].end());
      }

      if (isfork){
	for (set<GRCNode *>::iterator j = dpsnodes_set[*i].begin();
	     j != dpsnodes_set[*i].end(); j++){
	  dpsnodes_vec[n].push_back(*j);
	  dpsnodes_brnumb[n].push_back(succnumb);
	}
      }

      succnumb++;
    }

    if (isfork){
      orderChildren(n, &dpsnodes_vec[n], &dpsnodes_brnumb[n]);
    }
    
    if ((n->dataSuccessors.size()>0)||(n->dataPredecessors.size()>0))
      dpsnodes_set[n].insert(n);

    if(debug) 
      print_dpsinfo(dpsnodes_set[n], dpsnodes_vec[n], dpsnodes_brnumb[n], refmap[n]);
    return;
  }

  void orderChildren(GRCNode *n, vector<GRCNode *> *dpsnodes, vector<int> *brnumbs)
  {
    vector<GRCNode *>::iterator i,k;
    set<GRCNode *>::iterator j;
    int x, y, sz, min1, min2, min3;
    map<int, int> br_prior_prim; //primary priority - defined by input-output dps btw bruches
    map<int, int> br_prior_sec; //secondary priority - defined by parrent-children constrain
    map<int, int> br_prior_third; //third priority - defined by arbitary in-output dps
    vector<int> ordered; //ordered brunch sequence
    vector<int>::iterator it;
    map<int, GRCNode *> conn;
    bool all_equal;

    assert(dpsnodes->size() == brnumbs->size());
    if (!n)
      return;
    if (n->successors.size()<2)
      return;
    
    sz = n->successors.size();
    for (x = 0; x < sz; x++){
      br_prior_prim[x] = 0;
      br_prior_sec[x] = 0;
      br_prior_third[x] = 0;
      conn[x] = n->successors[x];
    }
    if(debug) cerr<<"FOR "<<refmap[n]<<"\n";
    sz = dpsnodes->size();
    for (x = 0; x < sz; x++){
      
      GRCNode *dpsn = (*dpsnodes)[x];
      int br_numb = (*brnumbs)[x];
      if(debug) cerr<<"dpsn "<<refmap[dpsn]<<" under brunch "<<br_numb<<":\n";
      //compute prim && third priority
      for (i = dpsn->dataSuccessors.begin();
	   i != dpsn->dataSuccessors.end(); i++){
	if (dpsnodes_set[n].count(*i) >0){
	  br_prior_prim[br_numb]++;
	}
	br_prior_third[br_numb]++;
      }

      for (i = dpsn->dataPredecessors.begin();
	   i != dpsn->dataPredecessors.end(); i++){
	if (dpsnodes_set[n].count(*i) >0){
	  br_prior_prim[br_numb]--;
	}
	br_prior_third[br_numb]--;
      }

      //compute secondary priority
      if (dpsn->dataPredecessors.size()>0){
	int a = 0, b = 0;
	for(j = dpsnodes_set[dpsn].begin(); j != dpsnodes_set[dpsn].end(); j++){
	  for (k = (*j)->dataSuccessors.begin();
	       k != (*j)->dataSuccessors.end(); k++){
	    //if belong to diff brunches, add up
	    if (dpsnodes_set[n].count(*k)>0)
	      if (findBrNumb((*k), dpsnodes, brnumbs) != br_numb)
		a++;
	  }
	}
	for (k = dpsn->dataPredecessors.begin();
	     k != dpsn->dataPredecessors.end(); k++){
	  if (dpsnodes_set[n].count(*k) > 0)
	    if (findBrNumb((*k),dpsnodes,brnumbs) != br_numb)
	      b++;
	}
	br_prior_sec[br_numb] += b * a;
      }
    }

    if (debug){      
      cerr<<"Node "<<refmap[n]<<"------------\n";
      for (x = 0; x < (int)n->successors.size(); x++){
	cerr<<"child "<<x<<"'s prim prior = "<<br_prior_prim[x]<<", "
	    <<"second prior = "<<br_prior_sec[x]<<"\n";
      }
    }

    all_equal = true;
    min1 = br_prior_prim[0];
    min2 = br_prior_sec[0];
    min3 = br_prior_third[0];
    //sz = br_prior_prim.size();
    sz = n->successors.size();
    for (x = 1; x < sz; x++)
      if ((br_prior_prim[x] != min1) || (br_prior_sec[x] != min2)
	  || (br_prior_third[x] != min3)){
	all_equal = false;
	break;
      }

    //if all brunches have the same primal && secondary && third priority, return
    if (all_equal)
      return;

    n->successors.clear();
    
    for (x = 0; x < sz; x++){
      ordered.push_back(x);
    }

    for (x = sz; x > 0; x--){
      it = ordered.begin();
      for (y = 0; y < x-1; y++){
	if ((br_prior_prim[y] > br_prior_prim[y+1])
	    ||
	    ((br_prior_sec[y] < br_prior_sec[y+1]) && (br_prior_prim[y] == br_prior_prim[y+1]))
	    ||
	    ((br_prior_sec[y] == br_prior_sec[y+1]) && (br_prior_prim[y] == br_prior_prim[y+1])
	     && (br_prior_third[y] > br_prior_third[y+1]))
	    ){
	  //exchange y && y+1, ie. it && it+1
	  int tmp = *it;
	  it = ordered.erase(it);
	  ordered.insert(it+1,tmp);
	  tmp = br_prior_prim[y];
	  br_prior_prim[y] = br_prior_prim[y+1];
	  br_prior_prim[y+1] = tmp;
	  tmp = br_prior_sec[y];
	  br_prior_sec[y] = br_prior_sec[y+1];
	  br_prior_sec[y+1] = tmp;
	  tmp = br_prior_third[y];
	  br_prior_third[y] = br_prior_third[y+1];
	  br_prior_third[y+1] = tmp;
	}
	it++;
      }
    }

    for (it = ordered.begin(); it != ordered.end(); it++){
      n->successors.push_back(conn[*it]);
    }

    if(debug) 
      print_ordered_ch(n);

  }

  int findBrNumb(GRCNode *n, vector<GRCNode *> *dpsnodes, vector<int> *brnumb)
  {
    for (int i = 0; i < (int)dpsnodes->size(); i++){
      if ((*dpsnodes)[i] == n)
	return (*brnumb)[i];
    }
    cerr<<"ERROR: Can't find "<<refmap[n]<<"\n";
    assert(0);
    return -1;
  }

  void print_dpsinfo(set<GRCNode *> dps, vector<GRCNode *> dps_ord, 
		     vector<int> br_numb, int node_numb)
  {
    set<GRCNode *>::iterator i;
    int x,sz;

    if(dps.size()==0)
      return;
    cerr<<"for node "<<node_numb<<"--------\n";
    cerr<<"dps nodes under are: ";
    for (i = dps.begin(); i != dps.end(); i++)
      cerr<<refmap[*i]<<" ";
    cerr<<"\n";

    sz = dps_ord.size();
    assert(dps_ord.size() == br_numb.size());
    if(sz == 0)
      return;
    cerr<<"dps in order: \n";
    for (x = 0; x < sz; x++)
      cerr<<refmap[dps_ord[x]]<<"-"<<br_numb[x]<<" ";
    cerr<<"\n";
  }

  void print_ordered_ch(GRCNode *n)
  {
    vector<GRCNode *>::iterator i;

    assert(n);
    cerr<<"Successors of "<<refmap[n]<<" after ordering:\n";
    for (i = n->successors.begin(); i != n->successors.end(); i++)
      cerr<<refmap[(*i)]<<" ";
    cerr<<"\n";
  }


};

int main(int argc, char *argv[])
{

  if(argc<4) 
    {
      cerr<<"Usage: ... pdg_file cpdg_file cpdg_dot_file\n"; return -1;}

  ifstream pdgfile(argv[1]); if(!pdgfile.good()) return -1;
  ofstream cpdg_file(argv[2]); if(!cpdg_file.good()) return -1;
  ofstream cpdg_dot_file(argv[3]); if(!cpdg_dot_file.good()) return -1;

  try {

    IR::XMListream r(pdgfile);
    IR::Node *n;
    r >> n;
    pdgfile.close();

    AST::Modules *mods = dynamic_cast<AST::Modules*>(n);
    if (!mods) throw IR::Error("Root node is not a Modules object");

    for ( vector<AST::Module*>::iterator i = mods->modules.begin() ;
	  i != mods->modules.end() ; i++ ) {
      assert(*i);

      GRCgraph *gf = dynamic_cast<GRCgraph*>((*i)->body);
      assert(gf);  
      GRCNode *top = gf->control_flow_graph;

      AST::SymbolTable *sig = dynamic_cast<AST::SymbolTable*>((*i)->signals);
      assert(sig);

      CFGmap refmap;
      STmap strefmap;
      
      int mxnode = gf->enumerate(refmap,strefmap);

      PDG2CCFG dpscut(top, sig, refmap);
      dpscut.excute();

      //mxnode = GRCDot::GRCDot(cpdg_dot_file, gf,(*i),false, false, refmap, strefmap, mxnode,true);
      mxnode = GRCDot::GRCDot(cpdg_dot_file, gf,(*i),false, false, refmap, strefmap, mxnode);

      cpdg_dot_file.close();
    }
    
    IR::XMLostream o(cpdg_file);
    o << n;
    cpdg_file.close();
    
  } catch (IR::Error &e) {
    cerr << e.s << endl;
    exit(-1);
  }

  return 0;

}
